/**
 * Author: Dave Toth
 * Date: Fall 2005
 */

import java.net.*;      // Use sockets.
import java.io.*;       // Read and write over sockets.
import java.util.*;     // Timers, Vectors, Hashtable
import java.util.Timer; // Timers
import java.awt.*;      // Timers

////////////////////////////////////////////////////////////////////
// DOWNLOAD THREAD
////////////////////////////////////////////////////////////////////

/**
 * Downloads a file from the web server (over http) to use for the 
 * work unit.  This method was adapted from "Java Network Programming" by 
 * Elliotte Rusty Harold page 87 and the Java API docs from Sun.  
 */

public class DownloadThread extends Thread {
    
    
    //////////////////////////////////////////////////////////////////////
    //                        Variables                                 //
    //////////////////////////////////////////////////////////////////////
    
    /*
     * Indicates whether the client should continue running.
     */
    boolean clientRunning = true;
    
    /**
     * The number of work units the client has completed.
     */
    int workUnitsDone = 0;
    
    /**
     * Used to write to sockets.
     */
    private PrintWriter writer;
    
    /**
     * Used to read from sockets.
     */
    private BufferedReader reader;
    
    /**
     * The Socket to use to connect to the server.
     */
    private Socket connection;
    
    /**
     * The unique ID of the client which is the last octet of its IP Address.
     */
    private String clientID;
    
    /**
     * The unique ID of the workunit the client is processing.
     */
    private String workUnitID;
    
    /**
     * The amount of time a client gets to complete a task before it is deemed late.
     */
    private long delayBound;
    
    /**
     * The name of the file to download.
     */
    private String fileName;
    
    /**
     * The Science Application Thread.
     */
    private ScienceApplication parent;
    
    /**
     * Indicates whether a thread shoul be paused (true) or killed (false).
     */
    public boolean threadPaused;
    
    
    
    
    /**
     * Constructor.
     */
    public DownloadThread(ScienceApplication parent, String clientID, long delayBound) {
	this.parent = parent;
	this.clientID = clientID;
	this.delayBound = delayBound;
    }
    
    public void run() {
	//System.out.println("%%%%%%%%%%%%%%%%%%%%%%%% DL TH started at: " + ValidatorClient.getCurrentSimulationTime());
	String workUnitName = "";
	try {
	    threadPaused = true;
	    String inputLine;
	    StringBuffer data = new StringBuffer();
	    
	    // Get the name of the work unit to process.
	    String returnValue = getWork(1, 1); // TODO should these be changed from 1 to something else?  Not yet.
	    workUnitID = returnValue.substring(0, returnValue.indexOf("_"));
	    workUnitName = returnValue.substring(1 + returnValue.indexOf("_"));
	    //System.out.println("File: " + workUnitName);
	    //System.out.println("ID: " + workUnitID);
	    
	    // Download the file you need to do the work unit.
	    URL url = new URL("http://192.168.2.21/data_files/" + workUnitName);
	    BufferedReader inputStream = new BufferedReader(new InputStreamReader(url.openStream()));
	    while ((inputLine = inputStream.readLine())!= null) {
		data.append(inputLine);
		data.append("\n");
	    }
	    
	    File f = new File(workUnitName);
	    if (f.exists() == true) {
		f.delete();
	    }

	    // Now create the file on the local disk.
	    FileWriter fileWriter = new FileWriter(workUnitName);
	    fileWriter.write(data.toString());
	    fileWriter.flush();
	    fileWriter.close();
	    
	    //System.out.println("data file downloaded OK at: " + ValidatorClient.getCurrentSimulationTime());
	}
	catch (Exception e) {
	    System.out.println("Error downloading" + e.toString());
	}
	if ((this.isInterrupted() == true) && (threadPaused == false)) {
	    //System.out.println("Download was Interrupted");
	    ValidatorClient.threadVector.remove(this);
	}
	else {
	    //System.out.println("Download Completed Successfully");
	    // Put the downloaded file into the buffer.
	    //System.out.println("delay bound" + delayBound);
	    long tempTime = System.currentTimeMillis();
	    parent.bufferWorkunit(new Task(workUnitName, workUnitID, tempTime + (delayBound * 1000)));
	    //System.out.println("Task created at " + tempTime + "  abort at " + (tempTime + (delayBound * 1000)) + "    cur time: " + (tempTime-ValidatorClient.startTime)/1000);
	    boolean ackResult = acknowledgeDownload();
	    ValidatorClient.threadVector.remove(this);
	}
    }
    
    
    /** 
     * Establish a connection to the server and return the result.
     * @return Returns OK if successful, FAILED otherwise.
     */
    public boolean connectToServer() {
	//System.out.println("in cnct to svr");
	try {
	    connection = new Socket("192.168.2.21", 19021);
	    writer = new PrintWriter(connection.getOutputStream());
	    reader = new BufferedReader(new InputStreamReader(connection.getInputStream()));
	    return true;
	}
	catch (Exception e) {
	    System.out.println("failed to connect to server");
	    return false;
	}
    }
    
    
    /**
     * Disconnect from the server.
     */
    public void disconnectFromServer() {
	try {
	    writer.close();
	    reader.close();
	    connection.close();
	}
	catch (Exception e) {
	    // Do nothing.  We're probably already disconnected.
	}
    } 
    
    
    /**
     * Sends a message to the server requesting work to do and gets a file
     * name back indicating what file will be required to do the work.
     * @return - The name of the file used for the work unit.
     */
    public String getWork(int fileSize, int computationSize) {
	//System.out.println("in get work");
	if (connectToServer() == true) {
	    writer.println(ValidatorClient.REQUEST_WORK + "-" + clientID + " " + fileSize + " " + computationSize);
	    writer.flush();
	    String message;
	    try {
		message = reader.readLine();
	    }
	    catch (IOException ioe) {
		message = "NO_WORK";
	    }
	    disconnectFromServer();
	    return message.trim();
	}
	else {
	    return "NO_WORK";
	}
    }
    
    
    /**
     * Signals server that it successfully downloaded a workunit.
     * @return - Returns the status of sending the acknowledgement to the server. 
     */
    public boolean acknowledgeDownload() {
	//System.out.println("Sending ack of download");
	if (connectToServer()) {
	    writer.write(ValidatorClient.ACKNOWLEDGE_DOWNLOAD + "-" + clientID + "-" + workUnitID + "=" + delayBound + "\n");
	    disconnectFromServer();
	    return true;
	}
	return false;
    }
}
